from .base import BaseDataset
from .alex_mi import AlexMI
from .bnci import BNCI2014001, BNCI2014004
from .tsinghua import Wang2016, BETA
from .nakanishi2015 import Nakanishi2015
from .zhou2016 import Zhou2016
from .physionet import PhysionetMI, PhysionetME
from .cho2017 import Cho2017
from .munich2009 import MunichMI
from .schirrmeister2017 import Schirrmeister2017
from .tunerl import Weibo2014
from .cbcic import CBCIC2019001, CBCIC2019004
